/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * Copyright (c) 2001 Martin Kutschker.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

const CALC_ADD = 0;
const CALC_SUB = 1;
const CALC_MUL = 2;
const CALC_DIV = 3;
const CALC_MOD = 4;
const CALC_POW = 5;
const CALC_ROT = 6;

const CALC_OPPRECED = [ 0, 0, 1, 1, 1, 2, 2 ]; // precedence of operators
const CALC_OPS = [ '+', '-', '*', '/', '%', '^', '!^' ];

function Calculator(mathClass, numberClass) {
  this.math   = mathClass || null;
  this.number = numberClass || null;

  this.operands    = new Array(0);
  this.operators   = new Array(0);

  this.reset     = reset;
  this.calculate = calculate;
  this.calcTop   = calcTop;

  this.evalFunction = evalFunction;

  this.parseRPN  = parseRPN;

  // reset the stack
  function reset() {
    this.operands.length  = 0;
    this.operators.length = 0;
  }

  // calculates the expression
  function calculate() {
    while(this.operators.length >= 1) {
      this.calcTop();
    }
  }

  // calculates top level operator of the stack
  function calcTop() {
    // sanity check
    if (this.operands.length < 2 || this.operators.length < 1) {
      alert("error - at least two operands and one operator expected: " +
            this.operands.length + " operand(s), " +
            this.operators.length + " operator(s)");
      this.clearAll();

      return;
    }

    var op2 = this.operands.pop();
    var op1 = this.operands.pop();

    switch (this.operators.pop()) {
      case CALC_ADD:
        op1.add(op2);
        break;

      case CALC_SUB:
        op1.sub(op2);
        break;
 
      case CALC_MUL:
        op1.mul(op2);
        break;

      case CALC_DIV:
        op1.div(op2);
        break;

      case CALC_MOD:
        op1.mod(op2);
        break;

      case CALC_POW:
        op1 = new this.number(this.math.pow(op1,op2));
        break;

      case CALC_ROT:
        // ugly hack!!!
        var tmp = "" + op2;
        var idx = tmp.indexOf(".");
        if (idx != -1 && tmp.substring(x+1) == "0")
          idx = -1;

        var negRes = false;

        if ((op1 < 0) &&                  // negative base
          idx == -1 &&                    // integer exponent
          this.number.mod(op2,2) == 1) {  // odd exponent
          op1.neg(); // should be abs()!
          negRes = true;
        }

        op2 = this.number.div(1, op2);
        op1 = new this.number(this.math.pow(op1,op2));

        if (negRes)
          op1.neg();
        break;

      default:
        op1 = null;
    }

    this.operands.push(op1);
  }

  function evalFunction(aFunc,op) {
    var val;
    var y;

    switch(aFunc) {
      case "exp":
        return (new this.number(this.math[aFunc](op)));

      // logarithmic functions
      case "ln": // logarithmus naturalis
        return (new this.number(this.math.log(op)));

      case "lg": // logarithmus
        return (new this.number.div(this.math.log(op),this.math.LN10));

      case "ld": // logarithmus dualis
        return (new this.number.div(this.math.log(op),this.math.LN2));

      // trigonometric functions
      case "sin":
      case "cos":
      case "tan":
        // convert to radians
        y = this.number.mul(op,this.math.PI);
        y.div(180);

        return (new this.number(this.math[aFunc](y)));

      case "asin":
      case "acos":
      case "atan":
        y = new this.number(this.math[aFunc](op));

        // convert to degree
        y.mul(180);
        y.div(this.math.PI);

        return (y);

      // hyperbolic functions
      case "sinh":
      case "cosh":
      case "tanh":
        var ex  = new this.number(this.math.exp(op));
        var enx = new this.number(this.math.exp(op.neg()));

        switch(aFunc) {
          case "sinh":
            ex.sub(enx);
            return (ex.div(2));

          case "cosh":
            ex.add(enx);
            return (ex.div(2));

          case "tanh":
            return (this.number.div(this.number.sub(ex,enx),
                                    this.number.add(ex,enx)));
        }

      case "asinh":
        y = this.number.mul(op,op);
        y = new this.number(this.math.sqrt(y.add(1)));
        y = new this.number(this.math.log(y.add(op)));
        return (y);

      case "acosh": // x >= 1
        y = this.number.mul(op,op);
        y = new this.number(this.math.sqrt(y.sub(1)));
        y = new this.number(this.math.log(y.add(op)));
        return (y);

      case "atanh": // -1 < x < 1
        y = this.number.add(1,op);
        y.div(this.number.sub(1,op));
        y = new this.number(this.math.log(y));
        y.div(2);
        return (y);

      case "fac": // x >= 0; x in N
        if (op < 0)
          return Number.NaN;

        // if (!op.isInt()) return return Number.NaN;

        y = new this.number(1);
        for (var i=new this.number(1); i<=op; i.add(1)) {
          y.mul(i);
        }
        return (y);
    }

    return (this.math.NaN);
  }

  // parse and evaluate expression in reverse polish notation
  function parseRPN(aRPN) {
    var ops = aRPN.split(" ");

    for (var i; i<ops.length; i++) {
      switch (ops[i]) {
        case "+":
          operators.push(CALC_ADD);
          this.calcTop();
          break;

        case "-":
          operators.push(CALC_SUB);
          this.calcTop();
          break;

        case "*":
          operators.push(CALC_MUL);
          this.calcTop();
          break;

        case "/":
          operators.push(CALC_DIV);
          this.calcTop();
          break;

        case "^":
          operators.push(CALC_POW);
          this.calcTop();
          break;

        default:
          operands.push(parseFloat(ops[i]));
      }
    }
  }

}
