
/*
 * A wrapper class around Number which implements
 * the arbitrary precision interface
 * and the binary math interface:
 *
 * XNumber.add(op1, op2) returns new XNumber(op1 + op2).
 * x.add(op1, op2) returns new XNumber(op1 + op2).
 * x.add(op1) adds op1 to the value of x and returns x.
 *
 * sub(), mul(), div(), mod(), and(), or(), xor() work similar.
 *
 * XNnumber.neg(op1) returns new XNumber(-op1)
 * x.neg(op1) returns new XNumber(-op1)
 * x.neg(1) negates the value of x and returns x
 *
 * not() works similar.
 *
 * Legal values for op are Number, String, XNumber or any object which
 * valueOf() or toNumber() methods returns Number.
 */

function XNumber(n) {
  this.value = XNumber.convert(n);

  this.valueOf  = XNumber_valueOf;
  this.toString = XNumber_toString;

  // missing:
  //
  // int() and fract() (Java: intValue, toBigInteger)
  //
  // isLT, ieEq, isGT and also isLToE, isGToE
  // (Java: compareTo ->  -1, 0, 1 = less than, equal, greather than)
  //
  // max(), min() (do weed them here and not in Math?)

  this.abs = XNumber_abs; // absolute value
  this.sig = XNumber_sig; // signum

  this.add = XNumber_add; // add
  this.sub = XNumber_sub; // subtract
  this.mul = XNumber_mul; // multiply
  this.div = XNumber_div; // divide
  this.mod = XNumber_mod; // modulo
  this.neg = XNumber_neg; // negate

  this.and = XNumber_and; // binary and
  this.or  = XNumber_or;  // binary or
  this.xor = XNumber_xor; // binary xor
  this.not = XNumber_not; // binary not
}

/***************************/

// converts argument to a Number (private static method)
function XNumber_convert(x) {
  switch (typeof(x)) {
    case "number":
      return x;

    case "string":
      return parseFloat(x);

    case "object":
      if (x.constructor == XNumber) {
        return x.value;
      }
      else {
        // see, if we can get a Number
        var xx = x.valueOf();
        if (typeof(xx) == "number")
          return xx;

        if (x.toNumber && typeof(x) == "function") {
          xx = x.toNumber();
          if (typeof(xx) == "number")
            return xx;
        }
      }
  }

  return Number.NaN;
}
XNumber.convert = XNumber_convert;

/***************************/

// create a new XNumber from string (static method)
function XNumber_parse(str) {
  return  new XNumber(parseFloat(str));
}
XNumber.parse = XNumber_parse;

/***************************/

// converts object to primitive
// (so 'x = new Xnumber(4); y = 3+x-Math.sin(x)' works)
function XNumber_valueOf() {
  return this.value;
}

/***************************/

// converts object to string
function XNumber_toString() {
  return this.value.toString();
}

/***************************/

function XNumber_abs(op1) {
  var val = Number.NaN;

  if (typeof(this) == "object"                // instance
      && arguments.length == 1) { // and no argument
    this.value = Math.abs(this.value);
    val        = this;
  }
  else
    val = new XNumber(Math.abs(XNumber.convert(op1)));

  return val;
}
XNumber.abs = XNumber_abs;

/***************************/

function XNumber_sig(op1) {
  var v1 = Number.NaN;

  if (typeof(this) == "object"                // instance
      && arguments.length == 1) { // and no argument
    v1 = this.value;
  }
  else
    v1 = XNumber.convert(op1);

  if (v1 < 0)
    return -1;

  if (v1 == 0)
    return 0;

  if (v1 > 0)
    return 1;

  return Number.NaN;
}
XNumber.sig = XNumber_sig;

/***************************/

function XNumber_add(op1,op2) {
  var val = Number.NaN;
  var v1  = XNumber.convert(op1);

  if (typeof(this) == "object"                // instance
      && arguments.length == 1) { // and single argument
    this.value += v1;
    val        = this;
  }
  else { // class or two arguments
    var v2 = XNumber.convert(op2);

    val = new XNumber(v1 + v2);
  }

  return val;
}
XNumber.add = XNumber_add;

/***************************/

function XNumber_sub(op1,op2) {
  var val = Number.NaN;
  var v1  = XNumber.convert(op1);

  if (typeof(this) == "object"                // instance
      && arguments.length == 1) { // and single argument
    this.value -= v1;
    val        = this;
  }
  else { // class or two arguments
    var v2 = XNumber.convert(op2);

    val = new XNumber(v1 - v2);
  }

  return val;
}
XNumber.sub = XNumber_sub;

/***************************/

function XNumber_mul(op1,op2) {
  var val = Number.NaN;
  var v1  = XNumber.convert(op1);

  if (typeof(this) == "object"                // instance
      && arguments.length == 1) { // and single argument
    this.value *= v1;
    val        = this;
  }
  else { // class or two arguments
    var v2 = XNumber.convert(op2);

    val = new XNumber(v1 * v2);
  }

  return val;
}
XNumber.mul = XNumber_mul;

/***************************/

function XNumber_div(op1,op2) {
  var val = Number.NaN;
  var v1  = XNumber.convert(op1);

  if (typeof(this) == "object"                // instance
      && arguments.length == 1) { // and single argument
    this.value /= v1;
    val        = this;
  }
  else { // class or two arguments
    var v2 = XNumber.convert(op2);

    val = new XNumber(v1 / v2);
  }

  return val;
}
XNumber.div = XNumber_div;

/***************************/

function XNumber_mod(op1,op2) {
  var val = Number.NaN;
  var v1  = XNumber.convert(op1);

  if (typeof(this) == "object"                // instance
      && arguments.length == 1) { // and single argument
    this.value %= v1;
    val        = this;
  }
  else { // class or two arguments
    var v2 = XNumber.convert(op2);

    val = new XNumber(v1 % v2);
  }

  return val;
}
XNumber.mod = XNumber_mod;

/***************************/

function XNumber_neg(op1) {
  var val = Number.NaN;

  if (typeof(this) == "object"                // instance
      && arguments.length == 0) { // and no argument
    this.value = -this.value;
    val        = this;
  }
  else { // class or one arguments
    val = new XNumber(-XNumber.convert(op1));
  }

  return val;
}
XNumber.neg = XNumber_neg;

/***************************/

function XNumber_and(op1,op2) {
  var val = Number.NaN;
  var v1  = XNumber.convert(op1);

  if (typeof(this) == "object"                // instance
      && arguments.length == 1) { // and single argument
    this.value &= v1;
    val        = this;
  }
  else { // class or two arguments
    var v2 = XNumber.convert(op2);

    val = new XNumber(v1 & v2);
  }

  return val;
}
XNumber.and = XNumber_and;

/***************************/

function XNumber_or(op1,op2) {
  var val = Number.NaN;

  if (typeof(this) == "object"                // instance
      && arguments.length == 1) { // and single argument
    this.value |= v1;
    val        = this;
  }
  else { // class or two arguments
    val = new XNumber(v1 | v2);
  }

  return val;
}
XNumber.or = XNumber_or;

/***************************/

function XNumber_xor(op1,op2) {
  var val = Number.NaN;
  var v1  = XNumber.convert(op1);

  if (typeof(this) == "object"                // instance
      && arguments.length == 1) { // and single argument
    this.value ^= v1;
    val        = this;
  }
  else { // class or two arguments
    var v2 = XNumber.convert(op2);

    val = new XNumber(v1 ^ v2);
  }

  return val;
}
XNumber.xor = XNumber_xor;

/***************************/

function XNumber_not(op1) {
  var val = Number.NaN;

  if (typeof(this) == "object"                // instance
      && arguments.length == 0) { // and no argument
    this.value = ~this.value;
    val        = this;
  }
  else { // class or one argument
    val = new XNumber(~v1);
  }

  return val;
}
XNumber.not = XNumber_not;
